@@ -3,13 +3,24 @@
 
 int main(int argc, char** argv)
 { XMethodsQuery query;
-  ArrayOfIDNamePair services;
   printf("Content-type: text/html\r\n\r\n<html><h1>Most Recent Xmethods Service Listing</h1><pre>\n");
-  if (!query.ns__getAllServiceNames(services))
-    services.print();
+  if (argc >= 2 && argv[1][1] == 's')
+  { ArrayOfServiceSummary services;
+    if (!query.ns__getAllServiceSummaries(services))
+      services.print();
+    else
+    { soap_print_fault(query.soap, stderr);
+      soap_print_fault_location(query.soap, stderr);
+    }
+  }
   else
-  { soap_print_fault(query.soap, stderr);
-    soap_print_fault_location(query.soap, stderr);
+  { ArrayOfIDNamePair services;
+    if (!query.ns__getAllServiceNames(services))
+      services.print();
+    else
+    { soap_print_fault(query.soap, stderr);
+      soap_print_fault_location(query.soap, stderr);
+    }
   }
   printf("</pre></html>\n");
   return 0;
