@@ -6,7 +6,7 @@ WSDL 1.1 binding schema implementation
 
 --------------------------------------------------------------------------------
 gSOAP XML Web services tools
-Copyright (C) 2001-2004, Robert van Engelen, Genivia, Inc. All Rights Reserved.
+Copyright (C) 2001-2005, Robert van Engelen, Genivia Inc. All Rights Reserved.
 This software is released under one of the following two licenses:
 GPL or Genivia's license for commercial use.
 --------------------------------------------------------------------------------
@@ -35,10 +35,6 @@ A commercial use license is available from Genivia, Inc., contact@genivia.com
 
 #include "wsdlH.h"
 
-#ifdef WITH_NONAMESPACES
-extern struct Namespace namespaces[];
-#endif
-
 using namespace std;
 
 const char *qname_token(const char *QName, const char *URI)
@@ -53,7 +49,20 @@ const char *qname_token(const char *QName, const char *URI)
 }
 
 int is_builtin_qname(const char *QName)
-{ return iflag || (QName ? *QName != '"' : 0); // if the QName does not start with a ", it must be in the nsmap
+{ if (iflag)
+    return 1;
+  if (QName)
+  { if (*QName != '"')
+      return 1;	// if the QName does not start with a ", it must be in the nsmap
+    const char *s = strchr(QName + 1, '"');
+    if (s)
+    { size_t n = s - QName - 1;
+      for (SetOfString::const_iterator i = exturis.begin(); i != exturis.end(); ++i)
+        if (strlen(*i) == n && !strncmp(QName + 1, *i, n))
+          return 1; // QName is in exturis
+    }
+  }
+  return 0;
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -67,6 +76,9 @@ int show_ignore(struct soap*, const char*);
 
 wsdl__definitions::wsdl__definitions()
 { soap = soap_new1(SOAP_XML_TREE | SOAP_C_UTFSTRING);
+#ifdef WITH_OPENSSL
+  soap_ssl_client_context(soap, SOAP_SSL_NO_AUTHENTICATION, NULL, NULL, NULL, NULL, NULL);
+#endif
 #ifdef WITH_NONAMESPACES
   soap_set_namespaces(soap, namespaces);
 #endif
@@ -82,6 +94,8 @@ wsdl__definitions::wsdl__definitions()
   targetNamespace = "";
   documentation = NULL;
   types = NULL;
+  updated = false;
+  redirs = 0;
 }
 
 wsdl__definitions::wsdl__definitions(struct soap *copy, const char *location)
@@ -93,49 +107,103 @@ wsdl__definitions::wsdl__definitions(struct soap *copy, const char *location)
   soap_default(soap);
   soap->fignore = warn_ignore;
   soap->encodingStyle = NULL;
+  updated = false;
+  redirs = 0;
   read(location);
 }
 
 wsdl__definitions::~wsdl__definitions()
-{ soap_done(soap);
+{ soap_destroy(soap);
+  soap_end(soap);
+  soap_done(soap);
   free(soap);
 }
 
 int wsdl__definitions::get(struct soap *soap)
-{ return traverse();
+{ return preprocess();
+}
+
+int wsdl__definitions::read(int num, char **loc)
+{ if (num <= 0)
+    return read(NULL);
+  if (num == 1)
+    return read(loc[0]);
+  wsdl__import im;
+  im.namespace_ = NULL;
+  for (int i = 0; i < num; i++)
+  { im.location = loc[i];
+    import.push_back(im);
+  }
+  return preprocess();
 }
 
 int wsdl__definitions::read(const char *location)
-{ if (location)
+{ char *local_path = NULL;
+  char save_host[sizeof(last_host)];
+  char save_path[sizeof(last_path)];
+  int save_port = last_port;
+  strcpy(save_host, last_host);
+  strcpy(save_path, last_path);
+  if (location)
   { if (!strncmp(location, "http://", 7) || !strncmp(location, "https://", 8))
-    { fprintf(stderr, "Connecting to '%s' to retrieve WSDL... ", location);
+    { fprintf(stderr, "Connecting to '%s' to retrieve WSDL/XSD... ", location);
       if (soap_connect_command(soap, SOAP_GET, location, NULL))
       { fprintf(stderr, "connection failed\n");
+        soap_print_fault(soap, stderr);
         exit(1);
       }
-      fprintf(stderr, "done\n");
+      fprintf(stderr, "connected, receiving...\n");
+      strcpy(last_host, soap->host);
+      strcpy(last_path, soap->path);
+      last_port = soap->port;
     }
-    else if (*soap->host)
-    { char *URL;
-      URL = strrchr(soap->path, '/');
-      if (URL)
-        *URL = '\0';
-      URL = (char*)soap_malloc(soap, strlen(soap->host) + strlen(soap->path) + strlen(location) + 32);
-      sprintf(URL, "http://%s:%d/%s/%s", soap->host, soap->port, soap->path, location);
-      fprintf(stderr, "Connecting to '%s' to retrieve '%s' WSDL... ", URL, location);
-      if (soap_connect_command(soap, SOAP_GET, URL, NULL))
+    else if (*last_host)
+    { local_path = strrchr(last_path, '/');
+      if (local_path)
+        *local_path = '\0';
+      local_path = (char*)soap_malloc(soap, strlen(last_host) + strlen(last_path) + strlen(location) + 32);
+      sprintf(local_path, "http://%s:%d%s/%s", last_host, last_port, last_path, location);
+      fprintf(stderr, "Connecting to '%s' to retrieve relative '%s' WSDL/XSD... ", local_path, location);
+      if (soap_connect_command(soap, SOAP_GET, local_path, NULL))
       { fprintf(stderr, "connection failed\n");
         exit(1);
       }
-      fprintf(stderr, "done\n");
+      fprintf(stderr, "connected, receiving...\n");
+      strcpy(last_host, soap->host);
+      strcpy(last_path, soap->path);
+      last_port = soap->port;
     }
     else
     { fprintf(stderr, "Reading file '%s'\n", location);
       soap->recvfd = open(location, O_RDONLY, 0);
       if (soap->recvfd < 0)
-      { fprintf(stderr, "Cannot open '%s'\n", location);
-        exit(1);
+      { if (*last_path)
+        { local_path = strrchr(last_path, '/');
+          if (local_path)
+            *local_path = '\0';
+          local_path = (char*)soap_malloc(soap, strlen(last_path) + strlen(location) + 2);
+	  strcpy(local_path, last_path);
+	  strcat(local_path, "/");
+	  strcat(local_path, location);
+          fprintf(stderr, "Opening '%s'\n", local_path);
+          soap->recvfd = open(local_path, O_RDONLY, 0);
+        }
+        if (soap->recvfd < 0 && import_path)
+        { local_path = (char*)soap_malloc(soap, strlen(import_path) + strlen(location) + 2);
+	  strcpy(local_path, import_path);
+	  strcat(local_path, "/");
+	  strcat(local_path, location);
+          fprintf(stderr, "Opening '%s'\n", local_path);
+          soap->recvfd = open(local_path, O_RDONLY, 0);
+	}
+        if (soap->recvfd < 0)
+	{ fprintf(stderr, "Cannot open '%s'\n", location);
+          exit(1);
+        }
+        location = local_path;
       }
+      strncpy(last_path, location, sizeof(last_path));
+      last_path[sizeof(last_path) - 1] = '\0';
     }
   }
   if (!soap_begin_recv(soap))
@@ -146,9 +214,10 @@ int wsdl__definitions::read(const char *location)
     { soap_retry(soap);
       xs__schema *schema = new xs__schema(soap);
       schema->soap_in(soap, "xs:schema", NULL);
-      if (schema->error())
-      { soap_print_fault(schema->soap, stderr);
-        soap_print_fault_location(schema->soap, stderr);
+      if (soap->error)
+      { fprintf(stderr, "An error occurred while parsing WSDL or XSD from '%s'\n", location?location:"");
+        soap_print_fault(soap, stderr);
+        soap_print_fault_location(soap, stderr);
         exit(1);
       }
       name = NULL;
@@ -158,11 +227,16 @@ int wsdl__definitions::read(const char *location)
       types = new wsdl__types();
       types->documentation = NULL;
       types->xs__schema_.push_back(schema);
-      traverse();
     }
     else if ((soap->error >= 301 && soap->error <= 303) || soap->error == 307) // HTTP redirect, socket was closed
-    { fprintf(stderr, "Redirected to '%s'\n", soap->endpoint);
-      return read(soap_strdup(soap, soap->endpoint));
+    { int r = SOAP_ERR;
+      fprintf(stderr, "Redirected to '%s'\n", soap->endpoint);
+      if (redirs++ < 10)
+        r = read(soap_strdup(soap, soap->endpoint));
+      else
+        fprintf(stderr, "Max redirects exceeded\n");
+      redirs--;
+      return r;
     }
     else
     { fprintf(stderr, "An error occurred while parsing WSDL from '%s'\n", location?location:"");
@@ -171,6 +245,8 @@ int wsdl__definitions::read(const char *location)
       exit(1);
     }
   }
+  if (vflag)
+    fprintf(stderr, "Done reading '%s'\n", location);
   soap_end_recv(soap);
   if (soap->recvfd >= 0)
   { close(soap->recvfd);
@@ -178,39 +254,87 @@ int wsdl__definitions::read(const char *location)
   }
   else
     soap_closesock(soap);
+  strcpy(last_host, save_host);
+  strcpy(last_path, save_path);
+  last_port = save_port;
   return SOAP_OK;
 }
 
-int wsdl__definitions::traverse()
-{ if (vflag)
-    cerr << "wsdl definitions " << (name?name:"") << " " << (targetNamespace?targetNamespace:"") << endl;
-  if (!targetNamespace)
-  { if (vflag)
-      fprintf(stderr, "Warning: WSDL %s has no targetNamespace\n", name?name:"");
-    targetNamespace = "";
-  }
-  // process import first
-  for (vector<wsdl__import>::iterator im = import.begin(); im != import.end(); ++im)
-    (*im).traverse(*this);
-  // merge imported WSDLs
+int wsdl__definitions::preprocess()
+{ // process import
+  for (vector<wsdl__import>::iterator im1 = import.begin(); im1 != import.end(); ++im1)
+    (*im1).preprocess(*this);
+  // merge nested imported WSDLs into single import list
 again:
-  for (vector<wsdl__import>::iterator i = import.begin(); i != import.end(); ++i)
-  { if ((*i).definitionsPtr())
-    { for (vector<wsdl__import>::iterator j = (*i).definitionsPtr()->import.begin(); j != (*i).definitionsPtr()->import.end(); ++j)
+  for (vector<wsdl__import>::iterator im2 = import.begin(); im2 != import.end(); ++im2)
+  { if ((*im2).definitionsPtr())
+    { for (vector<wsdl__import>::iterator i = (*im2).definitionsPtr()->import.begin(); i != (*im2).definitionsPtr()->import.end(); ++i)
       { bool found = false;
-        for (vector<wsdl__import>::iterator k = import.begin(); k != import.end(); ++k)
-        { if ((*j).definitionsPtr() && (*k).definitionsPtr() && !strcmp((*j).definitionsPtr()->targetNamespace, (*k).definitionsPtr()->targetNamespace))
+        for (vector<wsdl__import>::iterator j = import.begin(); j != import.end(); ++j)
+        { if (((*i).definitionsPtr() == (*j).definitionsPtr())
+	   || ((*i).location && (*j).location && !strcmp((*i).location, (*j).location)))
           { found = true;
 	    break;
 	  }
         }
         if (!found)
-        { import.push_back(*j);
+        { if (vflag)
+	    cerr << "Adding imported WSDL " << ((*i).location?(*i).location:"") << " to " << (name?name:"") << " namespace " << (targetNamespace?targetNamespace:"") << endl;
+	  import.push_back(*i);
           goto again;
         }
       }
     }
   }
+  // merge <types>
+  for (vector<wsdl__import>::iterator im3 = import.begin(); im3 != import.end(); ++im3)
+  { if ((*im3).definitionsPtr() && (*im3).definitionsPtr()->types)
+    { if (!types)
+      { types = soap_new_wsdl__types(soap, -1);
+        types->soap_default(soap);
+      }
+      // merge <types>, check for duplicates, add namespaces for sloppy imports
+      for (vector<xs__schema*>::const_iterator i = (*im3).definitionsPtr()->types->xs__schema_.begin(); i != (*im3).definitionsPtr()->types->xs__schema_.end(); ++i)
+      { bool found = false;
+	if (!(*i)->targetNamespace)
+          cerr << "Warning: Schema without namespace, assigning " << (targetNamespace?targetNamespace:"") << endl;
+	if (!(*i)->targetNamespace)
+	  (*i)->targetNamespace = targetNamespace;
+	else
+	{ for (vector<xs__schema*>::const_iterator j = types->xs__schema_.begin(); j != types->xs__schema_.end(); ++j)
+	  { if ((*j)->targetNamespace && !strcmp((*i)->targetNamespace, (*j)->targetNamespace))
+	    { found = true;
+	      break;
+	    }
+	  }
+	}
+	if (!found)
+        { if (vflag)
+	    cerr << "Adding schema " << ((*i)->targetNamespace?(*i)->targetNamespace:"") << " to types in WSDL " << (name?name:"") << " namespace " << (targetNamespace?targetNamespace:"") << endl;
+	  types->xs__schema_.push_back(*i);
+	}
+      }
+    }
+  }
+  if (types)
+    types->preprocess(*this);
+  return SOAP_OK;
+}
+
+int wsdl__definitions::traverse()
+{ if (vflag)
+    cerr << "wsdl definitions '" << (name?name:"") << "' targetNamespace " << (targetNamespace?targetNamespace:"?") << endl;
+  if (updated)
+    return SOAP_OK;
+  updated = true;
+  if (!targetNamespace)
+  { if (vflag)
+      fprintf(stderr, "Warning: WSDL '%s' has no targetNamespace\n", name?name:"");
+    targetNamespace = "";
+  }
+  // process import first
+  for (vector<wsdl__import>::iterator im = import.begin(); im != import.end(); ++im)
+    (*im).traverse(*this);
   // then process the types
   if (types)
     types->traverse(*this);
@@ -227,7 +351,7 @@ again:
   for (vector<wsdl__service>::iterator sv = service.begin(); sv != service.end(); ++sv)
     (*sv).traverse(*this);
   if (vflag)
-    cerr << "end of wsdl definitions " << (name?name:"") << " " << (targetNamespace?targetNamespace:"") << endl;
+    cerr << "end of wsdl definitions '" << (name?name:"") << "' targetNamespace " << (targetNamespace?targetNamespace:"?") << endl;
   return SOAP_OK;
 }
 
@@ -292,9 +416,6 @@ wsdl__port::wsdl__port()
 { bindingRef = NULL;
 }
 
-wsdl__port::~wsdl__port()
-{ }
-
 int wsdl__port::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl port" << endl;
@@ -330,7 +451,7 @@ int wsdl__port::traverse(wsdl__definitions& definitions)
     }
   }
   if (!bindingRef)
-    cerr << "Warning: could not find port " << (name?name:"") << " binding " << (binding?binding:"") << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+    cerr << "Warning: no port '" << (name?name:"") << "' binding '" << (binding?binding:"") << "' in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
   return SOAP_OK;
 }
 
@@ -346,9 +467,6 @@ wsdl__binding::wsdl__binding()
 { portTypeRef = NULL;
 }
 
-wsdl__binding::~wsdl__binding()
-{ }
-
 int wsdl__binding::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl binding" << endl;
@@ -383,7 +501,7 @@ int wsdl__binding::traverse(wsdl__definitions& definitions)
     }
   }
   if (!portTypeRef)
-    cerr << "Warning: could not find binding " << (name?name:"") << " portType " << (type?type:"") << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+    cerr << "Warning: no binding '" << (name?name:"") << "' portType '" << (type?type:"") << "' in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
   for (vector<wsdl__binding_operation>::iterator i = operation.begin(); i != operation.end(); ++i)
     (*i).traverse(definitions, portTypeRef);
   return SOAP_OK;
@@ -401,9 +519,6 @@ wsdl__binding_operation::wsdl__binding_operation()
 { operationRef = NULL;
 }
 
-wsdl__binding_operation::~wsdl__binding_operation()
-{ }
-
 int wsdl__binding_operation::traverse(wsdl__definitions& definitions, wsdl__portType *portTypeRef)
 { if (vflag)
     cerr << "wsdl binding operation" << endl;
@@ -414,18 +529,18 @@ int wsdl__binding_operation::traverse(wsdl__definitions& definitions, wsdl__port
   for (vector<wsdl__ext_fault>::iterator i = fault.begin(); i != fault.end(); ++i)
     (*i).traverse(definitions);
   operationRef = NULL;
-  if (portTypeRef)
+  if (name && portTypeRef)
   { for (vector<wsdl__operation>::iterator i = portTypeRef->operation.begin(); i != portTypeRef->operation.end(); ++i)
     { if ((*i).name && !strcmp((*i).name, name))
       { operationRef = &(*i);
         if (vflag)
-	  cerr << "Found operation " << (name?name:"") << endl;
+          cerr << "Found operation " << (name?name:"") << endl;
         break;
       }
     }
   }
   if (!operationRef)
-    cerr << "Warning: could not find operation " << (name?name:"") << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+    cerr << "Warning: no operation '" << (name?name:"") << "' in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
   else
   { for (vector<wsdl__ext_fault>::iterator i = fault.begin(); i != fault.end(); ++i)
     { if ((*i).name)
@@ -449,7 +564,7 @@ int wsdl__binding_operation::traverse(wsdl__definitions& definitions, wsdl__port
         }
       }
       if (!(*i).messagePtr())
-        cerr << "Warning: could not find soap:fault message in WSDL definitions " << (definitions.name?definitions.name:"") << " operation " << (name?name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+        cerr << "Warning: no soap:fault message in WSDL definitions '" << (definitions.name?definitions.name:"") << "' operation '" << (name?name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
     }
   }
   return SOAP_OK;
@@ -466,16 +581,20 @@ wsdl__operation *wsdl__binding_operation::operationPtr() const
 int wsdl__ext_input::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl ext input" << endl;
-  for (vector<soap__header>::iterator i = soap__header_.begin(); i != soap__header_.end(); ++i)
-    (*i).traverse(definitions);
+  for (vector<soap__header>::iterator hd = soap__header_.begin(); hd != soap__header_.end(); ++hd)
+    (*hd).traverse(definitions);
+  if (mime__multipartRelated_)
+    mime__multipartRelated_->traverse(definitions);
   return SOAP_OK;
 }
 
 int wsdl__ext_output::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl ext output" << endl;
-  for (vector<soap__header>::iterator i = soap__header_.begin(); i != soap__header_.end(); ++i)
-    (*i).traverse(definitions);
+  for (vector<soap__header>::iterator hd = soap__header_.begin(); hd != soap__header_.end(); ++hd)
+    (*hd).traverse(definitions);
+  if (mime__multipartRelated_)
+    mime__multipartRelated_->traverse(definitions);
   return SOAP_OK;
 }
 
@@ -483,9 +602,6 @@ wsdl__ext_fault::wsdl__ext_fault()
 { messageRef = NULL;
 }
 
-wsdl__ext_fault::~wsdl__ext_fault()
-{ }
-
 int wsdl__ext_fault::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl ext fault" << endl;
@@ -525,9 +641,6 @@ wsdl__input::wsdl__input()
 { messageRef = NULL;
 }
 
-wsdl__input::~wsdl__input()
-{ }
-
 int wsdl__input::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl input" << endl;
@@ -562,7 +675,7 @@ int wsdl__input::traverse(wsdl__definitions& definitions)
     }
   }
   if (!messageRef)
-    cerr << "Warning: could not find input " << (name?name:"") << " message " << (message?message:"") << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+    cerr << "Warning: no input '" << (name?name:"") << "' message '" << (message?message:"") << "' in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
   return SOAP_OK;
 }
 
@@ -578,9 +691,6 @@ wsdl__output::wsdl__output()
 { messageRef = NULL;
 }
 
-wsdl__output::~wsdl__output()
-{ }
-
 int wsdl__output::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl output" << endl;
@@ -615,7 +725,7 @@ int wsdl__output::traverse(wsdl__definitions& definitions)
     }
   }
   if (!messageRef)
-    cerr << "Warning: could not find output " << (name?name:"") << " message " << (message?message:"") << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+    cerr << "Warning: no output '" << (name?name:"") << "' message '" << (message?message:"") << "' in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
   return SOAP_OK;
 }
 
@@ -631,9 +741,6 @@ wsdl__fault::wsdl__fault()
 { messageRef = NULL;
 }
 
-wsdl__fault::~wsdl__fault()
-{ }
-
 int wsdl__fault::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl fault" << endl;
@@ -668,7 +775,7 @@ int wsdl__fault::traverse(wsdl__definitions& definitions)
     }
   }
   if (!messageRef)
-    cerr << "Warning: could not find fault " << (name?name:"") << " message " << (message?message:"") << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+    cerr << "Warning: no fault '" << (name?name:"") << "' message '" << (message?message:"") << "' in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
   return SOAP_OK;
 }
 
@@ -694,9 +801,6 @@ wsdl__part::wsdl__part()
   complexTypeRef = NULL;
 }
 
-wsdl__part::~wsdl__part()
-{ }
-
 int wsdl__part::traverse(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl part" << endl;
@@ -745,16 +849,16 @@ int wsdl__part::traverse(wsdl__definitions& definitions)
     { if (is_builtin_qname(element))
 	definitions.builtinElement(element);
       else
-        cerr << "Warning: could not find part " << (name?name:"") << " element " << element << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+        cerr << "Warning: no part '" << (name?name:"") << "' element '" << element << "' in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
     }
     else if (type)
     { if (is_builtin_qname(type))
 	definitions.builtinType(type);
       else
-        cerr << "Warning: could not find part " << (name?name:"") << " type " << type << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+        cerr << "Warning: no part '" << (name?name:"") << "' type '" << type << "' in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
     }
     else
-      cerr << "Warning: could not find part " << (name?name:"") << " element or type" << " in WSDL definitions " << (definitions.name?definitions.name:"") << " namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
+      cerr << "Warning: no part '" << (name?name:"") << "' element or type in WSDL definitions '" << (definitions.name?definitions.name:"") << "' namespace " << (definitions.targetNamespace?definitions.targetNamespace:"") << endl;
   }
   return SOAP_OK;
 }
@@ -783,80 +887,93 @@ xs__complexType *wsdl__part::complexTypePtr() const
 { return complexTypeRef;
 }
 
-int wsdl__types::traverse(wsdl__definitions& definitions)
+int wsdl__types::preprocess(wsdl__definitions& definitions)
 { if (vflag)
     cerr << "wsdl types" << endl;
-  // import external schemas
+  // link imported schemas, need to repeat when <types> is extended with new imported schema (from inside another schema, etc.)
 again:
   for (vector<xs__schema*>::iterator schema1 = xs__schema_.begin(); schema1 != xs__schema_.end(); ++schema1)
   { for (vector<xs__import>::iterator import = (*schema1)->import.begin(); import != (*schema1)->import.end(); ++import)
-    { if ((*import).namespace_ && (*import).schemaLocation)
-      { bool found = false;
-        for (vector<xs__schema*>::const_iterator schema2 = xs__schema_.begin(); schema2 != xs__schema_.end(); ++schema2)
+    { if ((*import).namespace_ && !(*import).schemaPtr())
+      { for (vector<xs__schema*>::const_iterator schema2 = xs__schema_.begin(); schema2 != xs__schema_.end(); ++schema2)
         { if ((*schema2)->targetNamespace && !strcmp((*import).namespace_, (*schema2)->targetNamespace))
-          { found = true;
+          { (*import).schemaPtr(*schema2);
 	    break;
 	  }
-        }
-	if (!found)
-	{ if ((*import).schemaPtr())
-	  { if (strcmp((*import).schemaPtr()->targetNamespace, (*import).namespace_))
-	      cerr << "Schema import namespace " << (*import).namespace_ << " does not correspond to imported targetNamespace " << (*import).schemaPtr()->targetNamespace << endl;
-	    else
-	    { xs__schema_.push_back((*import).schemaPtr());
-	      goto again;
-	    }
-	  }
-	  else
-          { struct Namespace *p = definitions.soap->local_namespaces;
-            const char *s = (*import).schemaLocation;
-            if (s && (*import).namespace_)
-            { if (p)
-              { for (; p->id; p++)
-                { if (p->in)
-                  { if (!soap_tag_cmp((*import).namespace_, p->in))
-                      break;
-	          }
-	          if (p->ns)
-                  { if (!soap_tag_cmp((*import).namespace_, p->ns))
-                      break;
-	          }
-                }
-              }
-	      else
-	        fprintf(stderr, "Warning: no namespace table\n");
-              if (!iflag && (!p || !p->id)) // don't import any of the schemas in the .nsmap table (or when -i option is used)
-	      { xs__schema *importschema = new xs__schema(definitions.soap, s);
-	        (*import).schemaPtr(importschema);
-	        if (!importschema->targetNamespace)
-	          importschema->targetNamespace = (*import).namespace_;
-	        if (strcmp(importschema->targetNamespace, (*import).namespace_))
-	          cerr << "Schema import namespace " << (*import).namespace_ << " does not correspond to imported targetNamespace " << importschema->targetNamespace << endl;
-	        else
-                { importschema->traverse();
-	          xs__schema_.push_back(importschema);
-	          goto again;
-	        }
-              }
-	    }
-	    else if (vflag)
-	      fprintf(stderr, "Warning: no schemaLocation for namespace '%s'\n", (*import).namespace_?(*import).namespace_:"");
-	  }
-        }
+	}
       }
     }
   }
+  // if a schema is imported but not in <types> then get it
   for (vector<xs__schema*>::iterator schema2 = xs__schema_.begin(); schema2 != xs__schema_.end(); ++schema2)
+  { for (vector<xs__import>::iterator import = (*schema2)->import.begin(); import != (*schema2)->import.end(); ++import)
+    { bool found = false;
+      if ((*import).namespace_)
+      { if ((*import).schemaPtr())
+	  found = true;
+        else
+	{ for (vector<xs__schema*>::const_iterator schema3 = xs__schema_.begin(); schema3 != xs__schema_.end(); ++schema3)
+          { if ((*schema3)->targetNamespace && !strcmp((*import).namespace_, (*schema3)->targetNamespace))
+            { found = true;
+	      break;
+	    }
+	  }
+	}
+	if (!found)
+	{ for (SetOfString::const_iterator i = exturis.begin(); i != exturis.end(); ++i)
+	  { if (!soap_tag_cmp((*import).namespace_, *i))
+            { found = true;
+	      break;
+            }
+          }
+        }
+      }
+      if (!found && !iflag) // don't import any of the schemas in the .nsmap table (or when -i option is used)
+      { xs__schema *importschema;
+        importschema = (*import).schemaPtr();
+	if (!importschema)
+        { const char *s = (*import).schemaLocation;
+	  if (!s)
+	    s = (*import).namespace_;
+	  importschema = new xs__schema(definitions.soap, s);
+	  if (!importschema->targetNamespace)
+	    importschema->targetNamespace = (*import).namespace_;
+	  else if ((*import).namespace_ && strcmp(importschema->targetNamespace, (*import).namespace_))
+	    cerr << "Schema import namespace " << ((*import).namespace_?(*import).namespace_:"") << " does not correspond to imported targetNamespace " << importschema->targetNamespace << endl;
+	}
+        for (vector<xs__schema*>::const_iterator schema3 = xs__schema_.begin(); schema3 != xs__schema_.end(); ++schema3)
+        { if ((*schema3)->targetNamespace && !strcmp((*import).namespace_, (*schema3)->targetNamespace))
+          { found = true;
+	    (*import).schemaPtr(*schema3);
+	    break;
+	  }
+	}
+	if (!found)
+	{ (*import).schemaPtr(importschema);
+	  xs__schema_.push_back(importschema);
+	  goto again;
+	}
+      }
+    }
+  }
+  return SOAP_OK;
+}
+
+int wsdl__types::traverse(wsdl__definitions& definitions)
+{ if (vflag)
+    cerr << "wsdl types" << endl;
+  for (vector<xs__schema*>::iterator schema3 = xs__schema_.begin(); schema3 != xs__schema_.end(); ++schema3)
   { // artificially extend the <import> of each schema to include others so when we traverse schemas we can resolve references
     for (vector<xs__schema*>::iterator importschema = xs__schema_.begin(); importschema != xs__schema_.end(); ++importschema)
-    { if (schema2 != importschema)
+    { if (schema3 != importschema && (*importschema)->targetNamespace)
       { xs__import *import = new xs__import();
         import->namespace_ = (*importschema)->targetNamespace;
         import->schemaPtr(*importschema);
-        (*schema2)->import.push_back(*import);
+        (*schema3)->import.push_back(*import);
       }
     }
-    for (vector<xs__import>::iterator import = (*schema2)->import.begin(); import != (*schema2)->import.end(); ++import)
+    // check and report
+    for (vector<xs__import>::iterator import = (*schema3)->import.begin(); import != (*schema3)->import.end(); ++import)
     { if ((*import).namespace_)
       { bool found = false;
         for (vector<xs__schema*>::const_iterator importschema = xs__schema_.begin(); importschema != xs__schema_.end(); ++importschema)
@@ -865,72 +982,73 @@ again:
 	    break;
 	  }
         }
-	if (!found)
+	if (!found && vflag)
 	  cerr << "Schema import namespace " << (*import).namespace_ << " refers to a known external Schema" << endl;
       }
       else
-        cerr << "<xs:import> has no namespace" << endl;
+        cerr << "Warning: Schema import " << ((*import).schemaLocation ? (*import).schemaLocation : "") << " has no namespace" << endl;
     }
   }
-  for (vector<xs__schema*>::iterator schema3 = xs__schema_.begin(); schema3 != xs__schema_.end(); ++schema3)
-  { (*schema3)->traverse();
-    if (vflag)
-      for (SetOfString::const_iterator i = (*schema3)->builtinTypes().begin(); i != (*schema3)->builtinTypes().end(); ++i)
+  // traverse the schemas
+  for (vector<xs__schema*>::iterator schema4 = xs__schema_.begin(); schema4 != xs__schema_.end(); ++schema4)
+    (*schema4)->traverse();
+  // find all built-in types, elements, and attributes
+  for (vector<xs__schema*>::iterator schema5 = xs__schema_.begin(); schema5 != xs__schema_.end(); ++schema5)
+  { if (vflag)
+      for (SetOfString::const_iterator i = (*schema5)->builtinTypes().begin(); i != (*schema5)->builtinTypes().end(); ++i)
         cerr << "Schema builtin type: " << (*i) << endl;
-    definitions.builtinTypes((*schema3)->builtinTypes());
-    definitions.builtinElements((*schema3)->builtinElements());
-    definitions.builtinAttributes((*schema3)->builtinAttributes());
+    definitions.builtinTypes((*schema5)->builtinTypes());
+    definitions.builtinElements((*schema5)->builtinElements());
+    definitions.builtinAttributes((*schema5)->builtinAttributes());
   }
   return SOAP_OK;
 }
 
+int wsdl__import::preprocess(wsdl__definitions& definitions)
+{ bool found = false;
+  if (vflag)
+    cerr << "preprocess wsdl import " << (location?location:"") << endl;
+  definitionsRef = NULL;
+  if (namespace_)
+  { for (SetOfString::const_iterator i = exturis.begin(); i != exturis.end(); ++i)
+    { if (!soap_tag_cmp(namespace_, *i))
+      { found = true;
+        break;
+      }
+    }
+  }
+  if (!found && location)
+  { // parse imported definitions
+    definitionsRef = new wsdl__definitions(definitions.soap, location);
+    if (!definitionsRef)
+      return SOAP_EOF;
+    if (!namespace_)
+      namespace_ = definitionsRef->targetNamespace;
+    else if (!definitionsRef->targetNamespace || !*definitionsRef->targetNamespace)
+      definitionsRef->targetNamespace = namespace_;
+    else if (strcmp(namespace_, definitionsRef->targetNamespace))
+      cerr << "Error: WSDL definitions/import " << location << " namespace " << namespace_ << " does not match imported targetNamespace " << definitionsRef->targetNamespace << endl;
+  }
+  else if (!found)
+    cerr << "WSDL definitions/import has no location attribute" << endl;
+  return SOAP_OK;
+}
+
 int wsdl__import::traverse(wsdl__definitions& definitions)
 { if (vflag)
-    cerr << "wsdl import" << endl;
-  definitionsRef = NULL;
-  if (location)
-  { // parse imported definitions
-    definitionsRef = new wsdl__definitions(definitions.soap, location);
-    if (!definitionsRef)
-      return SOAP_EOF;
-    if (namespace_)
-      if (strcmp(namespace_, definitionsRef->targetNamespace))
-        cerr << "Import " << location << " namespace " << namespace_ << " does not match imported targetNamespace " << definitionsRef->targetNamespace << endl;
-    // merge <types>
-    if (definitionsRef->types)
-    { if (definitions.types)
-      { // import schemas
-	/* Removed: now handled by xs__schema::traverse()
-        for (vector<xs__schema>::iterator schema = definitions.types->xs__schema_.begin(); schema != definitions.types->xs__schema_.end(); ++schema)
-        { for (vector<xs__schema>::iterator importschema = definitionsRef->types->xs__schema_.begin(); importschema != definitionsRef->types->xs__schema_.end(); ++importschema)
-	  { xs__import *import = new xs__import();
-	    if (vflag)
-	      cerr << "Importing schema " << (*importschema).targetNamespace << endl;
-	    import->namespace_ = (*importschema).targetNamespace; // the schemaRef pointer will be set when traversing <types>
-            (*schema).import.push_back(*import);
-	  }
-	}
-	*/
-        // merge <types>, check for duplicates
-        for (vector<xs__schema*>::const_iterator importschema = definitionsRef->types->xs__schema_.begin(); importschema != definitionsRef->types->xs__schema_.end(); ++importschema)
-	{ bool found = false;
-	  for (vector<xs__schema*>::const_iterator schema = definitions.types->xs__schema_.begin(); schema != definitions.types->xs__schema_.end(); ++schema)
-	  { if (!strcmp((*importschema)->targetNamespace, (*schema)->targetNamespace))
-	    { found = true;
-	      break;
-	    }
-	  }
-	  if (!found)
-	    definitions.types->xs__schema_.push_back(*importschema);
-	}
-      }
-      else
-        definitions.types = definitionsRef->types;
-    }
-    // combine sets of builtin types, elements, and attributes
-    definitions.builtinTypes(definitionsRef->builtinTypes());
-    definitions.builtinElements(definitionsRef->builtinElements());
-    definitions.builtinAttributes(definitionsRef->builtinAttributes());
+    cerr << "wsdl import " << (location?location:"") << endl;
+  if (definitionsRef)
+  { for (vector<wsdl__message>::iterator mg = definitionsRef->message.begin(); mg != definitionsRef->message.end(); ++mg)
+      (*mg).traverse(definitions);
+    // process portTypes before bindings
+    for (vector<wsdl__portType>::iterator pt = definitionsRef->portType.begin(); pt != definitionsRef->portType.end(); ++pt)
+      (*pt).traverse(definitions);
+    // process bindings
+    for (vector<wsdl__binding>::iterator bg = definitionsRef->binding.begin(); bg != definitionsRef->binding.end(); ++bg)
+      (*bg).traverse(definitions);
+    // process services
+    for (vector<wsdl__service>::iterator sv = definitionsRef->service.begin(); sv != definitionsRef->service.end(); ++sv)
+      (*sv).traverse(definitions);
   }
   return SOAP_OK;
 }
@@ -947,11 +1065,6 @@ wsdl__import::wsdl__import()
 { definitionsRef = NULL;
 }
 
-wsdl__import::~wsdl__import()
-{ // if (definitionsRef)
-    // delete definitionsRef;
-}
-
 ////////////////////////////////////////////////////////////////////////////////
 //
 //	streams
@@ -969,6 +1082,7 @@ ostream &operator<<(ostream &o, const wsdl__definitions &e)
     soap_begin_send(&soap);
     e.soap_out(&soap, "wsdl:definitions", 0, NULL);
     soap_end_send(&soap);
+    soap_destroy(&soap);
     soap_end(&soap);
     soap_done(&soap);
   }
